// types

struct float3
{
    float x, y, z;
};
typedef struct float3 float3;

struct PerRayData_radiance
{
    float3 result;
    float importance;
    int depth;
};
typedef struct PerRayData_radiance PerRayData_radiance;

struct Ray {
    float3 origin;
    float3 direction;
    unsigned int ray_type;
    float tmin;
    float tmax;
};
typedef struct Ray Ray;

enum RTtransformkind {
    RT_WORLD_TO_OBJECT = 0xf00,
    RT_OBJECT_TO_WORLD
};
typedef enum RTtransformkind RTtransformkind;

// function signatures

float dot(float3 a, float3 b);
float3 exp(float /*float3*/);
float fresnel_schlick(float cos_theta, float exponent, float minimum, float maximum);
float3 make_float3(float s);
int min(int a, int b);
float3 mul(float3 a, float3 b);
float3 mul13( float a, float3 b );
float3 normalize(float3 v);
float optix_cc_luminance( float3 rgb );
float3 plus_eq( float3 lhs, float3 rhs );
float3 reflect(float3 i, float3 n);
int refract(float3 r, float3 i, float3 n, float ior);
float3 rtTransformNormal( RTtransformkind kind, const float3 n );
float3 rtTransformPoint( RTtransformkind kind, const float3 p );
float3 TraceRay(float3, float3, int, float);

// global variables

extern float3 back_hit_point;
extern float3 cutoff_color;
extern float extinction_constant;
extern float fresnel_exponent;
extern float fresnel_maximum;
extern float fresnel_minimum;
extern float3 front_hit_point;
extern float importance_cutoff;
extern int max_depth;
extern PerRayData_radiance prd_radiance;
extern Ray ray;
extern float3 reflection_color;
extern int reflection_maxdepth;
extern float3 refraction_color;
extern float refraction_index;
extern int refraction_maxdepth;
extern float3 shading_normal;
extern float t_hit;

// -----------------------------------------------------------------------------

__attribute__((__kernelfun__)) float3 closest_hit_radiance_impl()
{

  const float3 n = normalize(rtTransformNormal(RT_OBJECT_TO_WORLD, shading_normal));
  const float3 fhp = rtTransformPoint(RT_OBJECT_TO_WORLD, front_hit_point);
  const float3 bhp = rtTransformPoint(RT_OBJECT_TO_WORLD, back_hit_point);
  const float3 i = ray.direction;
        float3 t;
        float3 r;

  float reflection = 1.0f;
  float3 result = make_float3(0.0f);

  const int depth = prd_radiance.depth;

  float3 beer_attenuation;
  if(dot(n, ray.direction) > 0) {

    beer_attenuation = exp(extinction_constant * t_hit);
  } else {
    beer_attenuation = make_float3(1);
  }


  if (depth < min(refraction_maxdepth, max_depth))
  {
    if ( refract(t, i, n, refraction_index) )
    {

      float cos_theta = dot(i, n);
      if (cos_theta < 0.0f)
        cos_theta = -cos_theta;
      else
        cos_theta = dot(t, n);

      reflection = fresnel_schlick(cos_theta, fresnel_exponent, fresnel_minimum, fresnel_maximum);

      float importance = prd_radiance.importance * (1.0f-reflection) * optix_cc_luminance( mul( refraction_color, beer_attenuation ) );
      float3 color = cutoff_color;
      if ( importance > importance_cutoff ) {
        color = TraceRay(bhp, t, depth+1, importance);
      }
      plus_eq( result, mul( mul13( (1.0f - reflection), refraction_color ), color ) );
    }

  }


  float3 color = cutoff_color;
  if (depth < min(reflection_maxdepth, max_depth))
  {
    r = reflect(i, n);

    float importance = prd_radiance.importance * reflection * optix_cc_luminance( mul( reflection_color, beer_attenuation ) );
    if ( importance > importance_cutoff ) {
      color = TraceRay( fhp, r, depth+1, importance );
    }
  }
  plus_eq( result, mul( mul13( reflection, reflection_color ), color ) );

  result = mul( result, beer_attenuation );

  return result;
}

