#!/bin/sh

# Usage: compile.sh source-name
#
# Compiles the source file (by compiling the entire project) and copies the
# PTX to the local directory.
#

quiet=false
if [ "X$1" = "X--quiet" ] ; then
    quiet=true
    shift
fi

if [ $# != 2 ] || [ ! -r "$2" ] ; then
    echo "Usage: $0 [--quiet] piece-file source-file" 1>&2
    exit 1
fi

root=`dirname "$0"`
root=`cd "$root" ; pwd`

# Identify the pieces of the program so they can be reassembled.

base=`echo "$1" | sed -e 's/_p[0-9][0-9]*.c$//'`
pieces=`echo ${base}_p* | tr " " "\n" | sed -e 's/\..*$//' | sort -u`

# Remove the program if it already exists. Then concatenate the pieces to
# rebuild it. For the piece we were asked to compile (i.e. $1) we need to
# de-mangle the C processed by GenProg into CUDA.

cufile=${base}.cu
rm -f $cufile
for piece in $pieces ; do
    if [ "X${piece}.c" = "X$1" ] ; then
        "$root/strip.py" $2 >> $cufile
    elif [ -r ${piece}.cu ] ; then
        cat ${piece}.cu >> $cufile
    fi
done

# Run make to compile the program and store the output stream to a log file
# for later processing. If we are not quiet, copy to both the log file and the
# screen. The nested subshell parens also allow us to capture the exit code and
# abort if make did not succeed.

if $quiet ; then
    make -C src > compile.log 2>&1 || exit $?
else
    ( exit `
        exec 3>&1 1>&4 4>&-
        ( make -C src 2>&1
        echo $? 1>&3
        ) | tee compile.log`
    ) 4>&1 || exit $?
fi

# Parse the compile log for variables that are used before being assigned. In
# practice, this eliminates a number of programs that are just going to produce
# bad images anyway.

cleanup()
{
    rm compile.log
    case $1 in
        0) exit 1 ;;
        *) exit 0 ;;
    esac
}

if $quiet ; then
    grep "used before its value is set" compile.log > /dev/null 2>&1
    cleanup $?
else
    ( exit `
        exec 3>&1 1>&4 4>&-
        ( grep "used before its value is set" compile.log
        echo $? 1>&3
        ) | awk '{ print "ERROR:" $0 }'`
    ) 4>&1
    cleanup $?
fi

