#!/usr/bin/python

# Usage: test.py [--quiet] [source-name fitness-file]
#
# Writes runtime and error measurements to the given fitness file. The source
# filename is used to determine the compiled ptx file's name; the md5sum of the
# ptx file is used for caching the results, since GenProg can often generate
# equivalent files that it thinks are different.
#
# If this script is called with no arguments, the cache and fitness file are
# skipped and the program is re-evaluated with the results sent to the screen.
#

from optparse import OptionParser
import os
import re
from subprocess import Popen, PIPE
import sys

parser = OptionParser( usage = "%prog [--quiet] [source-name fitness-file]" )
parser.add_option(
    "--quiet", action = "store_true", help = "do not print to screen"
)
parser.add_option(
    "-t", "--threshold", help = "optional threshold value to test"
)
options, args = parser.parse_args()

if len( args ) == 0:
    use_cache = False
elif len( args ) == 2:
    use_cache = True
    source_file  = args[ 0 ]
    fitness_file = args[ 1 ]
else:
    parser.print_help()
    exit( 1 )

tools = os.path.abspath( os.path.dirname( sys.argv[ 0 ] ) )

def run( *args ):
    if options.quiet:
        with open( "/dev/null", 'w' ) as stderr:
            return Popen( args, stdout=PIPE, stderr=stderr ).communicate()[ 0 ]
    else:
        return Popen( args, stdout=PIPE ).communicate()[ 0 ]

fitness = sys.stdout
if use_cache and fitness_file != "-":
    fitness = open( fitness_file, 'w' )

# If caching is enabled, determine the cache key (md5sum of the ptx file) and
# see if the results are already cached.

if use_cache:
    pat = re.compile( "_p\d+.c$" )
    hashfile = "ptx.cache"
    ptxfile  = pat.sub( ".ptx", source_file )

    md5sum = run( "md5sum", ptxfile ).split()[ 0 ]
    lines  = run( "grep", md5sum, hashfile )
    if len( lines ) > 0:
        md5sum, fname, time, error = lines.split( "\n" )[ 0 ].split()
        print >>fitness, time, error
        exit()

# Caching is disabled or the key was not found in the cache. Run the program 5
# times and record the run times. We report the minimum as the fitness value.

times = list()
errors = list()
for i in range( 5 ):
    cmd = [ "src/glass", "-A", "-B", "-s=test" ]
    if options.threshold is not None:
        cmd += [ "-t", options.threshold ]
    lines = run( *cmd )
    for line in lines.split( "\n" ):
        if line.startswith( "BENCHMARK" ):
            times.append( float( line.split()[ -2 ] ) )
            break
    else:
        continue

    # Compute the L2 difference between the golden image and the test image.
    # Save the total per-pixel error as the fitness.

    lines = run( os.path.join( tools, "diff-img" ), "golden.ppm", "test.ppm" )
    line = lines.split( "\n" )[ 0 ]
    errors.append( float( line.split()[ -1 ] ) )
    os.remove( "test.ppm" )

    if not options.quiet:
        print times[ -1 ], errors[ -1 ]
if len( times ) == 0:
    exit( 1 )

# Report the minimum time and average error values

time = min( times )
error = errors[ 0 ]
for i, val in enumerate( errors[ 1: ] ):
    error = error + ( val - error ) / ( i + 2 )

print >>fitness, time, error
if use_cache:
    f = open( hashfile, 'a' )
    print >>f, md5sum, ptxfile, time, error

